/*
 * Simple device driver to rotate encrypt strings. Useful for getting
 * an overview of the device driver API in the Linux kernel.
 *
 * It registers as a misc device and uses /dev, /sys and /proc entries
 * to communicate to the kernel.
 *
 * 2016, Razvan Deaconescu <razvan.deaconescu@cs.pub.ro>
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/kfifo.h>
#include <asm/uaccess.h>
#include <linux/ctype.h>
#include <linux/mutex.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>


MODULE_DESCRIPTION("Rotation Encryption Device Driver");
MODULE_AUTHOR("Razvan Deaconescu");
MODULE_LICENSE("GPL");


#define DEFAULT_ROT_OFFSET	13
static unsigned short rot_offset = DEFAULT_ROT_OFFSET;
module_param(rot_offset, ushort, S_IRUGO|S_IWUSR);
MODULE_PARM_DESC(rot_offset, "offset value to use for rot encryption; default is 13");

#define FIFO_SIZE		4096
static DEFINE_KFIFO(fifo, char, FIFO_SIZE);
static char tmp_buffer[FIFO_SIZE];
static DEFINE_MUTEX(fifo_mutex);

static int rot_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int rot_release(struct inode *inode, struct file *file)
{
	return 0;
}

static char do_rot(char v)
{
	unsigned short offset;

	/* Normalize offset. */
	offset = rot_offset % ('z' - 'a' + 1);

	if (islower(v)) {
		v += offset;
		if (islower(v))
			return v;
		else
			return v - 'z' + 'a';
	}
	else if (isupper(v)) {
		v += offset;
		if (isupper(v))
			return v;
		else
			return v - 'Z' + 'A';
	}
	return v;
}

static ssize_t rot_read(struct file *file,
		char __user *user_buffer,
		size_t size, loff_t *offset)
{
	size_t copied;
	size_t i;
	int err;

	mutex_lock(&fifo_mutex);
	copied = kfifo_out(&fifo, tmp_buffer, size);
	for (i = 0; i < copied; i++)
		tmp_buffer[i] = do_rot(tmp_buffer[i]);

	err = copy_to_user(user_buffer, tmp_buffer + *offset, copied);
	if (err != 0) {
		pr_err("rot: error copying to user space\n");
		return err;
	}
	mutex_unlock(&fifo_mutex);
	*offset += copied;

	return copied;
}

static ssize_t rot_write(struct file *file,
		const char __user *user_buffer,
		size_t size, loff_t *offset)
{
	size_t copied;
	int err;

	mutex_lock(&fifo_mutex);
	err = kfifo_from_user(&fifo, user_buffer + *offset, size, &copied);
	if (err != 0) {
		pr_err("rot: error copying from user space\n");
		return err;
	}
	*offset += copied;
	mutex_unlock(&fifo_mutex);

	return copied;
}

static struct file_operations rot_fops = {
	.owner = THIS_MODULE,
	.open = rot_open,
	.release = rot_release,
	.read = rot_read,
	.write = rot_write,
};

/*
static ssize_t rot_proc_read(struct file *file,
		char __user *user_buffer,
		size_t size, loff_t *offset)
{
	char buffer[32];
	size_t copied;
	int err;

	sprintf(buffer, "%hu\n", rot_offset);
	if (strlen(buffer) > size)
		copied = size;
	else
		copied = strlen(buffer);
	err = copy_to_user(user_buffer, buffer, copied);
	if (err != 0) {
		pr_err("rot: error copying to userspace");
		return -EFAULT;
	}
	*offset += copied;

	return copied;
}
*/

static int rot_proc_show(struct seq_file *m, void *v)
{
	seq_printf(m, "%hu\n", rot_offset);
	return 0;
}

static int rot_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, rot_proc_show, NULL);
}

static ssize_t rot_proc_write(struct file *file,
		const char __user *user_buffer,
		size_t size, loff_t *offset)
{
	unsigned long tmp_offset;
	char buffer[32];
	size_t copied;
	int err;

	if (31 > size)
		copied = size;
	else
		copied = 31;
	err = copy_from_user(buffer, user_buffer, copied);
	if (err != 0) {
		pr_err("rot: error copying from userspace\n");
		return -EFAULT;
	}
	*offset += copied;
	if (buffer[copied-1] == '\n')
		buffer[copied-1] = '\0';
	buffer[copied] = '\0';

	err = kstrtoul(buffer, 10, &tmp_offset);
	if (err != 0) {
		pr_err("rot: error converting string to number\n");
		return err;
	}
	/* Normalize offset. */
	rot_offset = tmp_offset % ('z' - 'a' + 1);

	return copied;
}

static struct file_operations rot_proc_fops = {
	.owner = THIS_MODULE,
	.open = rot_proc_open,
	.release = single_release,
	.read = seq_read,
	.write = rot_proc_write,
};

static struct proc_dir_entry *proc_dir, *proc_offset;

#define ROT_MINOR		42
static struct miscdevice rot_dev = {
	ROT_MINOR,
	"rot",
	&rot_fops
};

static int rot_init(void)
{
	int err;

	proc_dir = proc_mkdir("rot", NULL);
	if (proc_dir == NULL) {
		pr_err("rot: can not create proc folder\n");
		return -EBUSY;
	}
	proc_offset = proc_create("offset", 0644, proc_dir, &rot_proc_fops);
	if (proc_offset == NULL) {
		pr_err("rot: can not create proc file\n");
		return -EBUSY;
	}

	err = misc_register(&rot_dev);
	if (err != 0) {
		pr_err("rot: misc device register failed\n");
		return err;
	}

	return 0;
}

static void rot_exit(void)
{
	misc_deregister(&rot_dev);
	proc_remove(proc_dir);
}

module_init(rot_init);
module_exit(rot_exit);
